import argparse
import os
from .transcribe import Transcribe, TranscriptionOptions
from .models import Models
from .languages import LANGUAGES, TO_LANGUAGE_CODE


def optional_int(string):
    return None if string == "None" else int(string)


def str2bool(string):
    str2val = {"True": True, "False": False}
    if string in str2val:
        return str2val[string]
    else:
        raise ValueError(f"Expected one of {set(str2val.keys())}, got {string}")


def optional_float(string):
    return None if string == "None" else float(string)


def compression_ratio(text) -> float:
    text_bytes = text.encode("utf-8")
    return len(text_bytes) / len(zlib.compress(text_bytes))


def read_command_line():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "audio", nargs="+", type=str, help="audio file(s) to transcribe"
    )
    parser.add_argument(
        "--model",
        default="small",
        choices=Models().get_list(),
        help="name of the Whisper model to use",
    )
    parser.add_argument(
        "--output_dir",
        "-o",
        type=str,
        default=".",
        help="directory to save the outputs",
    )
    parser.add_argument(
        "--output_format",
        "-f",
        type=str,
        default="all",
        choices=["txt", "vtt", "srt", "tsv", "all"],
        help="format of the output file; if not specified, all available formats will be produced",
    )

    parser.add_argument(
        "--verbose",
        type=str2bool,
        default=True,
        help="whether to print out the progress and debug messages",
    )

    parser.add_argument(
        "--task",
        type=str,
        default="transcribe",
        choices=["transcribe", "translate"],
        help="whether to perform X->X speech recognition ('transcribe') or X->English translation ('translate')",
    )
    parser.add_argument(
        "--language",
        type=str,
        default=None,
        choices=sorted(LANGUAGES.keys())
        + sorted([k.title() for k in TO_LANGUAGE_CODE.keys()]),
        help="language spoken in the audio, specify None to perform language detection",
    )
    parser.add_argument(
        "--threads",
        type=optional_int,
        default=0,
        help="number of threads used for CPU inference",
    )

    #    parser.add_argument("--model_dir", type=str, default=None, help="the path to save model files; uses ~/.cache/whisper by default")
    parser.add_argument(
        "--temperature", type=float, default=0, help="temperature to use for sampling"
    )
    parser.add_argument(
        "--best_of",
        type=optional_int,
        default=5,
        help="number of candidates when sampling with non-zero temperature",
    )
    parser.add_argument(
        "--beam_size",
        type=optional_int,
        default=5,
        help="number of beams in beam search, only applicable when temperature is zero",
    )
    parser.add_argument(
        "--patience",
        type=float,
        default=1.0,
        help="optional patience value to use in beam decoding, as in https://arxiv.org/abs/2204.05424, the default (1.0) is equivalent to conventional beam search",
    )
    parser.add_argument(
        "--length_penalty",
        type=float,
        default=1.0,
        help="optional token length penalty coefficient (alpha) as in https://arxiv.org/abs/1609.08144, uses simple length normalization by default",
    )

    parser.add_argument(
        "--suppress_tokens",
        type=str,
        default="-1",
        help="comma-separated list of token ids to suppress during sampling; '-1' will suppress most special characters except common punctuations",
    )
    parser.add_argument(
        "--initial_prompt",
        type=str,
        default=None,
        help="optional text to provide as a prompt for the first window.",
    )
    parser.add_argument(
        "--condition_on_previous_text",
        type=str2bool,
        default=True,
        help="if True, provide the previous output of the model as a prompt for the next window; disabling may make the text inconsistent across windows, but the model becomes less prone to getting stuck in a failure loop",
    )
    #   parser.add_argument("--temperature_increment_on_fallback", type=optional_float, default=0.2, help="temperature to increase when falling back when the decoding fails to meet either of the thresholds below")
    parser.add_argument(
        "--compression_ratio_threshold",
        type=optional_float,
        default=2.4,
        help="if the gzip compression ratio is higher than this value, treat the decoding as failed",
    )
    parser.add_argument(
        "--logprob_threshold",
        type=optional_float,
        default=-1.0,
        help="if the average log probability is lower than this value, treat the decoding as failed",
    )
    parser.add_argument(
        "--no_speech_threshold",
        type=optional_float,
        default=0.6,
        help="if the probability of the <|nospeech|> token is higher than this value AND the decoding has failed due to `logprob_threshold`, consider the segment as silence",
    )
    parser.add_argument(
        "--word_timestamps",
        type=str2bool,
        default=False,
        help="(experimental) extract word-level timestamps and refine the results based on them",
    )
    #    parser.add_argument("--prepend_punctuations", type=str, default="\"\'“¿([{-", help="if word_timestamps is True, merge these punctuation symbols with the next word")
    # parser.add_argument("--append_punctuations", type=str, default="\"\'.。,，!！?？:：”)]}、", help="if word_timestamps is True, merge these punctuation symbols with the previous word"
    parser.add_argument(
        "--device", default="auto", help="device to use for CTranslate2 inference"
    )

    # CTranslate2 specific parameters
    parser.add_argument(
        "--compute_type",
        choices=["int8", "int8_float16", "int16", "float16"],
        default="int8",
        help="Type of quantization to use",
    )

    return parser.parse_args().__dict__


def main():
    args = read_command_line()
    output_dir: str = args.pop("output_dir")
    output_format: str = args.pop("output_format")
    os.makedirs(output_dir, exist_ok=True)
    model: str = args.pop("model")
    threads: int = args.pop("threads")
    language: str = args.pop("language")
    task: str = args.pop("task")
    device: str = args.pop("device")
    compute_type: str = args.pop("compute_type")
    verbose: bool = args.pop("verbose")

    options = TranscriptionOptions(
        beam_size=args.pop("beam_size"),
        best_of=args.pop("best_of"),
        patience=args.pop("patience"),
        length_penalty=args.pop("length_penalty"),
        log_prob_threshold=args.pop("logprob_threshold"),
        no_speech_threshold=args.pop("no_speech_threshold"),
        compression_ratio_threshold=args.pop("compression_ratio_threshold"),
        condition_on_previous_text=args.pop("condition_on_previous_text"),
        temperatures=[args.pop("temperature")],
        initial_prompt=args.pop("initial_prompt"),
        #        prefix = None,
        #        suppress_blank = None,
        suppress_tokens=args.pop("suppress_tokens"),
        #        without_timestamps = None,
        #        max_initial_timestamp =  None,
        word_timestamps=args.pop("word_timestamps"),
        #        prepend_punctuations = None,
        #        append_punctuations = None
    )

    model_dir = Models().get_model_dir(model)

    for audio_path in args.pop("audio"):
        Transcribe().inference(
            audio_path,
            model_dir,
            output_dir,
            output_format,
            task,
            language,
            threads,
            device,
            compute_type,
            verbose,
            options,
        )


if __name__ == "__main__":
    main()



#
# Based on code from https://github.com/openai/whisper
#
import hashlib
import os
import urllib.request
from tqdm import tqdm


class Models:
    _MODELS = {
        "tiny": "https://huggingface.co/jordimas/whisper-ct2/resolve/main/1e2aa48a43d45b9304a22aca730c5b5177a923f50008bffd26140e3b6ff19fdf.tiny/",
        "tiny.en": "https://huggingface.co/jordimas/whisper-ct2/resolve/main/d82547957fb5dffbe279d22292980ab3ad33675a231faadbc9f2c908b909b42d.tiny-en/",
        "base": "https://huggingface.co/jordimas/whisper-ct2/resolve/main/2b3a5459a344d5cfc6959018fb01f78ad610460d184ddf6a3282c187e7ddeceb.base/",
        "base.en": "https://huggingface.co/jordimas/whisper-ct2/resolve/main/67b9ae74793ba78292c0c11db4bf296169bbfbc8bd9f81dadf22546b9fa0e809.base-en/",
        "small": "https://huggingface.co/jordimas/whisper-ct2/resolve/main/64c8c26522e710cfad21c44c037a806cccf0dda756ce0a222b5f3a938d87de43.small/",
        "small.en": "https://huggingface.co/jordimas/whisper-ct2/resolve/main/ee51a5c0132eb205931bfdd635fae2ada72f5580c812ab113ca2d735ea2707e9.small-en/",
        "medium": "https://huggingface.co/jordimas/whisper-ct2/resolve/main/309bb6f6fc40dedef3526916b4110288796ca6a873ed73dfc9f3b7bbbf5ecccd.medium/",
        "medium.en": "https://huggingface.co/jordimas/whisper-ct2/resolve/main/4f2c05f42abebfd9f5f9027e26966153b38995722a066ee8da14e6dbdc976622.medium-en/",
        "large": "https://huggingface.co/jordimas/whisper-ct2/resolve/main/d2bb7bcdf195976460f6ba94f4222d3ba8d6b306d0a7d98447f3d58eecac4a28.large/",
    }

    _FILES = ["model.bin", "config.json", "vocabulary.txt"]

    def __init__(self):
        default = os.path.join(os.path.expanduser("~"), ".cache")
        self.download_root = os.path.join(
            os.getenv("XDG_CACHE_HOME", default), "whisper-ctranslate2"
        )

    def get_list(self) -> str:
        return list(self._MODELS.keys())

    def get_model_path(self, model) -> str:
        url = self._MODELS[model]
        model_dir = url.split("/")[-2]
        return os.path.join(self.download_root, model_dir)

    def get_model_dir(self, model):
        if model not in self._MODELS:
            raise RuntimeError(f"Model {model} not supported")

        model_path = self.get_model_path(model)

        if self._are_localfile_checksums_correct(model, self._MODELS[model]):
            return model_path

        for _file in self._FILES:
            self._download(model, self._MODELS[model] + _file)

        if self._are_localfile_checksums_correct(model, self._MODELS[model]):
            return model_path

        raise RuntimeError(
            "Model has been downloaded but the SHA256 checksum does not not match. Please retry downloading the model."
        )

    def _are_localfile_checksums_correct(self, model: str, url: str) -> bool:
        model_path = self.get_model_path(model)
        model_bytes = bytes()
        expected_sha256 = url.split("/")[-2].rsplit(".", 1)[0]

        for _file in self._FILES:
            _file = os.path.join(model_path, _file)

            if not os.path.exists(_file):
                return False

            with open(_file, "rb") as f:
                model_bytes += f.read()

        got = hashlib.sha256(model_bytes).hexdigest()
        return got == expected_sha256

    def _download(self, model: str, url: str) -> str:
        model_path = self.get_model_path(model)
        os.makedirs(model_path, exist_ok=True)

        expected_sha256 = url.split("/")[-2].rsplit(".", 1)[0]
        download_target = os.path.join(model_path, os.path.basename(url))

        if os.path.exists(download_target) and not os.path.isfile(download_target):
            raise RuntimeError(f"{download_target} exists and is not a regular file")

        with urllib.request.urlopen(url) as source, open(
            download_target, "wb"
        ) as output:
            with tqdm(
                total=int(source.info().get("Content-Length")),
                ncols=80,
                unit="iB",
                unit_scale=True,
                unit_divisor=1024,
            ) as loop:
                while True:
                    buffer = source.read(8192)
                    if not buffer:
                        break

                    output.write(buffer)
                    loop.update(len(buffer))



from .writers import get_writer, format_timestamp
from typing import NamedTuple, Optional, List
import tqdm
import sys

try:
    from faster_whisper import WhisperModel
except ModuleNotFoundError as e:
    print("Error faster_whisper dependency not found")
    print(
        "Install faster_whisper dependency by typing: pip install git+https://github.com/guillaumekln/faster-whisper"
    )
    exit()

system_encoding = sys.getdefaultencoding()

if system_encoding != "utf-8":

    def make_safe(string):
        return string.encode(system_encoding, errors="replace").decode(system_encoding)

else:

    def make_safe(string):
        return string


class TranscriptionOptions(NamedTuple):
    beam_size: int
    best_of: int
    patience: float
    length_penalty: float
    log_prob_threshold: Optional[float]
    no_speech_threshold: Optional[float]
    compression_ratio_threshold: Optional[float]
    condition_on_previous_text: bool
    temperatures: List[float]
    initial_prompt: Optional[str]
    #    prefix: Optional[str]
    #    suppress_blank: bool
    suppress_tokens: Optional[List[int]]
    #    without_timestamps: bool
    #    max_initial_timestamp: float
    word_timestamps: bool


#    prepend_punctuations: str
#    append_punctuations: str


class Transcribe:
    def inference(
        self,
        audio: str,
        model_path: str,
        output_dir: str,
        output_format: str,
        task: str,
        language: str,
        threads: int,
        device: str,
        compute_type: str,
        verbose: bool,
        options: TranscriptionOptions,
    ):
        model = WhisperModel(
            model_path, device=device, compute_type=compute_type, cpu_threads=threads
        )

        segments, info = model.transcribe(
            audio=audio,
            language=language,
            task=task,
            beam_size=options.beam_size,
            best_of=options.best_of,
            patience=options.patience,
            length_penalty=options.length_penalty,
            temperature=options.temperatures,
            compression_ratio_threshold=options.compression_ratio_threshold,
            log_prob_threshold=options.log_prob_threshold,
            no_speech_threshold=options.no_speech_threshold,
            condition_on_previous_text=options.condition_on_previous_text,
            initial_prompt=options.initial_prompt,
            # suppress_tokens = options.suppress_tokens,
            word_timestamps=options.word_timestamps,
        )

        print(
            "Detected language '%s' with probability %f"
            % (info.language, info.language_probability)
        )

        list_segments = []
        last_pos = 0
        accumated_inc = 0
        with tqdm.tqdm(
            total=info.duration, unit="seconds", disable=verbose is not False
        ) as pbar:
            for segment in segments:
                if verbose:
                    start, end, text = segment.start, segment.end, segment.text
                    line = f"[{format_timestamp(start)} --> {format_timestamp(end)}] {text}"
                    print(make_safe(line))

                list_segments.append(segment)
                duration = segment.end - last_pos
                increment = (
                    duration
                    if accumated_inc + duration < info.duration
                    else info.duration - accumated_inc
                )
                accumated_inc += increment
                last_pos = segment.end
                pbar.update(increment)

        writer = get_writer(output_format, output_dir)
        results = {}
        results["segments"] = list_segments
        writer(results, audio)
